/**********************************************************************
 * ISO MPEG Audio Subgroup Software Simulation Group (1996)
 * ISO 13818-3 MPEG-2 Audio Decoder - Lower Sampling Frequency Extension
 *
 * $Id: mp12_mps_tool.c,v 1.4 2008-03-05 12:42:34 sehlstlf Exp $
 *
 * $Log: mp12_mps_tool.c,v $
 * Revision 1.4  2008-03-05 12:42:34  sehlstlf
 * get the  last MPS frame for layer III after not finding a next sync and header
 *
 * Revision 1.3  2008/03/04 13:10:55  purnhaho
 * fixed long vs. int to avoid warning
 *
 * Revision 1.2  2008/03/04 08:25:08  sehlstlf
 * mps_file_name initialized to empty string
 *
 * Revision 1.1.1.1  2008/02/29 17:08:17  purnhaho
 * initial import of mp12_mps_tool.zip from Philips 2008-02-29
 *
 * Revision 1.22  2006/03/10 15:11:56  dodenhen
 * Added MPEG_2_5 support
 *
 * Revision 1.21  2004/10/04 14:08:45  kaehleof
 * added switch for alias reduction in mixed blocks
 *
 * Revision 1.20  2004/09/29 14:57:10  kaehleof
 * moved the decoder tables into tables.c and .h;
 * no more tables/ directory needed now;
 * create_tables.c was used to create tables.c and .h
 * from the files in the tables/ directory;
 *
 * Revision 1.19  2004/08/10 14:34:15  sps
 * added some code to allow for investigation of channel configuration
 * changes, to be enabled using the define CHANNEL_INVESTIGATION in
 * decoder.h
 *
 * Revision 1.18  2003/11/17 15:02:15  tmm01
 *   by Ernst F. Schroeder, look for: efs
 * + added code for decoding free format,
 *   especially in Layer III.
 *   The first sync word is examined, and when
 *   free format is detected, the actual number
 *   of slots is determined from the distance
 *   of the following sync word.
 *
 * Revision 1.17  2003/11/13 12:20:31  tmm01
 *   by Ernst F. Schroeder, look for: efs
 * + added routines for CRC check in Layer III
 *   CRC errors are displayed and counted -
 *   no further action is taken
 *
 * Revision 1.16  2003/10/01 13:40:27  sps
 * for raw pcm output: disabled decoder abort in the presence of an
 * implicit channel re-configuration
 *
 * Revision 1.15  2003/05/13 17:15:34  sps
 *  + bug fix in SubBandSynthesis() (previously introduced while changing
 *    to automatic arrays (was previously malloc))
 *
 *  + revision of the out_fifo interface (only the error code is returned,
 *    the clipping information became a function parameter)
 *
 *  + revised the code such that it can be compiled using g++
 *
 *  + functions returning an error code have now the return type
 *    ERROR_CODE. This concerns
 *     - out_fifo()
 *     - SubBandSynthesis()
 *     - III_get_side_info()
 *     - III_hufman_decode()
 *     - read_syn_window()
 *     - open_bit_stream_r()
 *     - initialize_huffman()
 *     - GetArguments()
 *     - aiff_read_headers()
 *     - aiff_seek_to_sound()
 *     - aiff_write_headers()
 *
 * Revision 1.14  2003/05/13 13:53:30  sps
 *  + bug fix in SubBandSynthesis() (previously introduced while changing
 *    to automatic arrays (was previously malloc)
 *
 *  + revision of the interface to out_fifo (only the error code is
 *    returned, the clipping information became a function parameter)
 *
 * Revision 1.13  2003/05/10 06:44:06  sps
 * bugfix - a bracket was set wrong, such that clipping was not output for L3
 *
 * Revision 1.12  2003/05/09 17:41:48  sps
 * +++ removed memory leaks.
 * +++ revised termination behaviour for unusual cases (file could not be
 * opened, no data could be written into file, memory allocation failed). A
 * couple of functions now return an error code, that leads to a clean
 * termination in the top level function.
 * +++ simplified legacy write functionality.
 * +++ removed some unnecessary "extern" declaration in function
 * declarations in header files
 *
 * Revision 1.11  2003/05/06 16:23:28  sps
 * +++ terminates if the number of channels in the bitstream exceeds the
 * number of initialized channels +++ allows to specify the number of
 * channels to be initialized using the command line option -n +++ performs
 * smart rounding from double to float prior to pass the data to the afsp
 * library write function +++ added function MyClose to assure that all
 * memory is free?d prior to exit
 *
 * Revision 1.10  2003/04/30 17:39:53  sps
 * +++ fixed aiff header for native output +++ fixed rounding for 16 bit
 * native output (by this, separated rounding in individual function
 * castDoubleToLong(), which is now called somewhat later (not anymore in
 * SubBandSynthesis(), but in out_fifo())) +++ made sure that the last
 * samples are written into the output file +++ made sure that decoding
 * starts only if file opening was successfull +++ added AFsp library
 * support, that allows the output of WAVE files. The output of raw and
 * aiff files is still supported, but is in this case also handled by the
 * AFsp library +++ Note: If the AFsp lib is used, any particular support
 * of erroneous signals is disabled +++
 *
 * Revision 1.9  2003/03/27 18:38:23  sps
 * Set pcm_aiff_data.sampleSize = 24, if 24 output is requested.
 *
 * Check whether file names and paths for input and outpt fit into the
 * assigned string arrays.
 *
 * Revision 1.8  2003/03/11 10:07:20  sps
 * committed contribution from Mike Coleman to remove certain kinds of
 * compiler warnings
 *
 * Revision 1.7  2002/11/21 18:35:26  sps
 * removed log message from header (it caused again a compiler warning)
 *
 * Revision 1.6  2002/11/21 18:31:23  sps
 *
 * Revision 1.5  2002/11/14 13:04:08  tmm01
 * warnings removed (Microsoft Visual Studio V.6.0)  23.10.2 W.Voessing
 *
 * Revision 1.2  1996/03/28 03:13:37  rowlands
 * Merged layers 1-2 and layer 3 revisions
 *
 * Revision 1.1  1996/02/14 03:45:52  rowlands
 * Initial revision
 *
 * Received from FhG
 **********************************************************************/
/**********************************************************************
 *   date   programmers                comment                        *
 * 2/25/91  Douglas Wong        start of version 1.0 records          *
 * 3/06/91  Douglas Wong        rename setup.h to dedef.h             *
 *                              removed extraneous variables          *
 *                              removed window_samples (now part of   *
 *                              filter_samples)                       *
 * 3/07/91  Davis Pan           changed output file to "codmusic"     *
 * 5/10/91  Vish (PRISM)        Ported to Macintosh and Unix.         *
 *                              Incorporated new "out_fifo()" which   *
 *                              writes out last incomplete buffer.    *
 *                              Incorporated all AIFF routines which  *
 *                              are also compatible with SUN.         *
 *                              Incorporated user interface for       *
 *                              specifying sound file names.          *
 *                              Also incorporated user interface for  *
 *                              writing AIFF compatible sound files.  *
 * 27jun91  dpwe (Aware)        Added musicout and &sample_frames as  *
 *                              args to out_fifo (were glob refs).    *
 *                              Used new 'frame_params' struct.       *
 *                              Clean,simplify, track clipped output  *
 *                              and total bits/frame received.        *
 * 7/10/91  Earle Jennings      changed to floats to FLOAT            *
 *10/ 1/91  S.I. Sudharsanan,   Ported to IBM AIX platform.           *
 *          Don H. Lee,                                               *
 *          Peter W. Farrett                                          *
 *10/ 3/91  Don H. Lee          implemented CRC-16 error protection   *
 *                              newly introduced functions are        *
 *                              buffer_CRC and recover_CRC_error      *
 *                              Additions and revisions are marked    *
 *                              with "dhl" for clarity                *
 * 2/11/92  W. Joseph Carter    Ported new code to Macintosh.  Most   *
 *                              important fixes involved changing     *
 *                              16-bit ints to long or unsigned in    *
 *                              bit alloc routines for quant of 65535 *
 *                              and passing proper function args.     *
 *                              Removed "Other Joint Stereo" option   *
 *                              and made bitrate be total channel     *
 *                              bitrate, irrespective of the mode.    *
 *                              Fixed many small bugs & reorganized.  *
 *19 aug 92 Soren H. Nielsen    Changed MS-DOS file name extensions.  *
 * 8/27/93 Seymour Shlien,      Fixes in Unix and MSDOS ports,        *
 *         Daniel Lauzon, and                                         *
 *         Bill Truerniet                                             *
 *--------------------------------------------------------------------*
 * 4/23/92  J. Pineda           Added code for layer III.  LayerIII   *
 *          Amit Gulati         decoding is currently performed in    *
 *                              two-passes for ease of sideinfo and   *
 *                              maindata buffering and decoding.      *
 *                              The second (computation) pass is      *
 *                              activated with "decode -3 <outfile>"  *
 * 10/25/92 Amit Gulati         Modified usage() for layerIII         *
 * 12/10/92 Amit Gulati         Changed processing order of re-order- *
 *                              -ing step.  Fixed adjustment of       *
 *                              main_data_end pointer to exclude      *
 *                              side information.                     *
 *  9/07/93 Toshiyuki Ishino    Integrated Layer III with Ver 3.9.    *
 *--------------------------------------------------------------------*
 * 11/20/93 Masahiro Iwadare    Integrated Layer III with Ver 4.0.    *
 *--------------------------------------------------------------------*
 *  7/14/94 Juergen Koller      Bug fixes in Layer III code           *
 *--------------------------------------------------------------------*
 * 08/11/94 IIS                 Bug fixes in Layer III code           *
 *--------------------------------------------------------------------*
 * 11/04/94 Jon Rowlands        Prototype fixes                       *
 *--------------------------------------------------------------------*
 *  7/12/95 Soeren H. Nielsen   Changes for LSF Layer I and II        *
 *--------------------------------------------------------------------*
 *  7/14/94 Juergen Koller      Bug fixes in Layer III code           *
 *--------------------------------------------------------------------*
 *     8/95 Roland Bitto        addapdet to MPEG 2                    *
 *--------------------------------------------------------------------*
 * 11/22/95 Heiko Purnhagen     skip ancillary data in bitstream      *
 *--------------------------------------------------------------------*
 * 05/08/02 Walter Voessing     new functions for MPEG1 and MPEG 2    *
 *                              Stereo Processing                     *  
 *                              bugs within 'III_get_scale_factors'   *   
 *                              fixed                                 *
 *                              'out_fifo': bugs fixed and 24 bit     *
 *                                          output added              *
 *                              'GetArguments': '-C' parameter added  *
 *                              (optional 24 bit output)              *
 *                              modified usage()                      *
 **********************************************************************/

#include        "common.h"
#include        "decoder.h"
#include        "huffman.h"

#ifdef USELIB_AFsp
#include <libtsp.h>
#include <libtsp/AFpar.h>
#include <libtsp/nucleus.h>
#endif

/* typedef short PCM[2][3][SBLIMIT]; */
/* typedef short PCM[2][SSLIMIT][SBLIMIT]; */
typedef double PCM[2][SSLIMIT][SBLIMIT];  /* Intensity Stereo Bug 06.11.01 Vo */
typedef unsigned int SAM[2][3][SBLIMIT];
typedef double FRA[2][3][SBLIMIT];
typedef double VE[2][HAN_SIZE];

/********************************************************************
 *
 *        This part contains the MPEG I decoder for Layers I & II.
 *
 *********************************************************************/

/****************************************************************
 *
 *        For MS-DOS user (Turbo c) change all instance of malloc
 *        to _farmalloc and free to _farfree. Compiler model hugh
 *        Also make sure all the pointer specified are changed to far.
 *
 *****************************************************************/
/*********************************************************************
 *
 * Core of the Layer II decoder.  Default layer is Layer II.
 *
 *********************************************************************/

/* Global variable definitions for "musicout.c" */

char *programName;

/* local functions */

static void MyClose(
              Bit_stream_struc  bs,
#ifdef USELIB_AFsp
              AFILE             *musicout,
#else
              FILE              *musicout,
#endif
              PCM FAR            pcm_sample,
              SAM FAR            sample,
              FRA FAR            fraction,
              VE  FAR            w
              )
{
/*   printf("\n\nreached MyClose\n\n"); */
  close_bit_stream_r(&bs);
  if ( musicout ) {
  }
  if (pcm_sample) free (pcm_sample);
  if (sample)     free (sample);
  if (fraction)   free (fraction);
  if (w)          free (w);
#ifdef USELIB_AFsp
  if (musicout) AFclose(musicout);
#else
  if (musicout) fclose(musicout);
#endif
}

static void MyExit(
              Bit_stream_struc  bs,
#ifdef USELIB_AFsp
              AFILE             *musicout,
#else
              FILE              *musicout,
#endif
              PCM FAR            pcm_sample,
              SAM FAR            sample,
              FRA FAR            fraction,
              VE  FAR            w,
              ERROR_CODE         returnCode

              )
{
    MyClose(bs,
            musicout,
            pcm_sample,
            sample,
            fraction,
            w);
    exit(returnCode);

}

#define CRC8_POLYNOMIAL     0x0007       /*7L*/
static void update_CRC8 (unsigned int data, unsigned int length,
                        unsigned int *crc, int logging)
{
  unsigned int masking, carry;

  masking = 1 << length;

  while ((masking >>= 1))
  {
    carry = *crc & 0x80;
    *crc <<= 1;
    if (!carry ^ !(data & masking))
      *crc ^= CRC8_POLYNOMIAL;
  }
  *crc &= 0xff;

  if (logging)
  {
    printf ("crc_len: %2d code: %02x crc: %02x\n", length, data, *crc);
    fflush (stdout);
  }
}

#define CRC16_POLYNOMIAL    0x8005   /*32773L*/
static void update_CRC16 (unsigned int data, unsigned int length,
                          unsigned int *crc, int logging)
{
  unsigned int masking, carry;

  masking = 1 << length;

  while ((masking >>= 1))
  {
    carry = *crc & 0x8000;
    *crc <<= 1;
    if (!carry ^ !(data & masking))
      *crc ^= CRC16_POLYNOMIAL;
  }
  *crc &= 0xffff;

  if (logging)
  {
    printf ("crc_len: %2d code: %04x crc: %04x\n", length, data, *crc);
    fflush (stdout);
  }
}

int find_mps_data (unsigned char *mpg_bytes, int nbytes, int frame_bytes, int *mps_pos, int *first, int *skip_bytes, int logging_on)
{
   int i, j;
   int mps_data_detected, mps_SSC_detected, crclen, curpos, cnt, crc_logging_on, new_skip_bytes = 0;
   int ancStart, ancStop;

   mps_data_detected = FALSE;
   mps_SSC_detected = FALSE;
   crc_logging_on = FALSE;
   /* check for presence of mps data from the start of the MPEG-1 ancillary data of the frame */
   for (i = frame_bytes; (i < nbytes-4) && !mps_data_detected; i++)
   {
      if ((mpg_bytes[i] == 0x8E) && ((mpg_bytes[i+1] & 0xE0) == 0x40))
      {
         if ((mpg_bytes[i+1] & 0x10) == 0x00)
            crclen = 1;
         else
            crclen = 2;

         curpos = i + 2;
         cnt = (int) mpg_bytes[curpos++];
         if (cnt == 255)
         {
            cnt += 256 * mpg_bytes[curpos] + mpg_bytes[curpos+1];
            curpos += 2;
         }
         if (curpos + crclen + cnt <= nbytes)
         {
            unsigned int crc_state, crc_received;

            if (crclen == 1)
            {
               crc_state = 0xff;
               crc_received = (unsigned int) mpg_bytes[curpos++];
               update_CRC8 ((unsigned int) (mpg_bytes[i+1] & 0x0F), 4, &crc_state, crc_logging_on);
               update_CRC8 ((unsigned int)  mpg_bytes[i+2],         8, &crc_state, crc_logging_on);
            }
            else
            {
               crc_state = 0xffff;
               crc_received = ((unsigned int) (mpg_bytes[curpos]) << 8) + (unsigned int) mpg_bytes[curpos+1];
               curpos += 2;
               update_CRC16 ((unsigned int) (mpg_bytes[i+1] & 0x0F), 4, &crc_state, crc_logging_on);
               update_CRC16 ((unsigned int)  mpg_bytes[i+2],         8, &crc_state, crc_logging_on);
            }
            if (cnt >= 255)
            {
               if (crclen == 1)
               {
                  update_CRC8 ((unsigned int) mpg_bytes[i+3], 8, &crc_state, crc_logging_on);
                  update_CRC8 ((unsigned int) mpg_bytes[i+4], 8, &crc_state, crc_logging_on);
               }
               else
               {
                  update_CRC16 ((unsigned int) mpg_bytes[i+3], 8, &crc_state, crc_logging_on);
                  update_CRC16 ((unsigned int) mpg_bytes[i+4], 8, &crc_state, crc_logging_on);
               }
            }
            for (j = 0; j < cnt; j++)
            {
               if (crclen == 1)
                  update_CRC8 ((unsigned int) mpg_bytes[curpos+j], 8, &crc_state, crc_logging_on);
               else
                  update_CRC16 ((unsigned int) mpg_bytes[curpos+j], 8, &crc_state, crc_logging_on);
            }
            if (crc_received != crc_state)
            {
               if (logging_on)
                  printf ("Error: CRC error\n");
               return 0;
            }
            else
            {
               mps_data_detected = TRUE;
               mps_SSC_detected = (mpg_bytes[i+1] & 0x04) == 0x04;
               ancStart = (mpg_bytes[i+1] & 0x02) == 0x02;
               ancStop = (mpg_bytes[i+1] & 0x01) == 0x01;
               if (mps_SSC_detected && ancStart)
               {
                  if (*first == 1)
                  {
                     *first = 0;
                  }
                  else
                  {
                     new_skip_bytes = mpg_bytes[curpos] + 1;
                  }
               }
            }
            if (logging_on)
               printf ("SSC: %d ancStart: %d ancStop: %d crc_received: %04x crc_calculated: %04x cnt: %3d\n",
                        mps_SSC_detected, ancStart, ancStop, crc_received, crc_state, cnt);
         }
         else
         {
            if (logging_on)
               printf ("Error: Spatial data (%d bytes) exceeds ancillary data (%d bytes).\n",
                        mpg_bytes[i+2]+4, nbytes-frame_bytes);
            return 0;
         }
      }
   }
   if (!mps_data_detected)
   {
      if (logging_on)
         printf ("Error: No AncDataElement detected.\n");
      return 0;
   }
   *mps_pos = curpos + *skip_bytes;
   cnt -= *skip_bytes;
   if (new_skip_bytes > cnt)
   {
      *skip_bytes = new_skip_bytes - cnt;
      new_skip_bytes = cnt;
   }
   else
   {
      *skip_bytes = 0;
   }
   *mps_pos += new_skip_bytes;

   return cnt - new_skip_bytes;
}

static void usage()
{
  printf("usage: %s [options] inputBS [outPCM] [outMPS]\n", programName);
  printf("options are:\n");
  printf(" -A       output is AIFF PCM\n");
  printf(" -R       output is RAW  PCM (big endian)");
#ifndef USELIB_AFsp
  printf(" (default)\n");
#else
  printf("\n");
  printf(" -W       output is WAVE PCM  (default)\n");
#endif
  printf(" -C       24 bit output (default is 16 bit)\n");		/* added 08.05.02 WV */
  printf(" -s sb    resynth only up to this sb (debugging only)\n");
  printf(" -n ch    initial number of channels (default 0)\n\
                    0 - according to first frame\n\
                    1 - mono\n\
                    2 - stereo\n");
  printf(" -a       antialias in mixed blocks turned off (ISO/IEC 11172-3 conform)\n");
  printf(" inputBS  input bit stream of encoded audio\n");
  printf(" outPCM   output PCM sound file (dflt inName+%s)\n", DFLT_OPEXT);
  printf(" outMPS   output MPEG Surround data file (dflt inName+%s)\n", DFLT_MPSEXT);
}


static ERROR_CODE GetArguments(
                               int          argc,
                               char       **argv,
                               Arguments_t *Arguments )
     
{
  int i=0, err=0;

  programName = argv[0];

  Arguments->Output24 = FALSE;   /* added to switch between 16 and 24 bit output WV 08.05.02 */
  Arguments->topSb = 0;
  Arguments->need_aiff = FALSE;
  Arguments->need_esps = FALSE;	/* MI */
  Arguments->need_raw  = FALSE;  
  Arguments->need_wav  = FALSE;  
  Arguments->nrOfChannels = 0;
  Arguments->antialias_mixed_blocks = 1;
  Arguments->encoded_file_name[0] = '\0';
  Arguments->decoded_file_name[0] = '\0';
  Arguments->mps_file_name[0] = '\0';

  /* interpret CL Args */
      
      
  while(++i<argc && err == 0) {
    char c, *token, *arg, *nextArg;
    int  argUsed;
        
    token = argv[i];
    if(*token++ == '-') {
      if(i+1 < argc) nextArg = argv[i+1];
      else           nextArg = "";
      argUsed = 0;
      while( (c = *token++) != 0) {
        if(*token /* NumericQ(token) */) arg = token;
        else                             arg = nextArg;
        switch(c) {
        case 's':  Arguments->topSb = atoi(arg); argUsed = 1;
          if(Arguments->topSb<1 || Arguments->topSb>SBLIMIT) {
            fprintf(stderr, "%s: -s band %s not %d..%d\n",
                    programName, arg, 1, SBLIMIT);
            err = 1;
          }
          break;
        case 'n':  Arguments->nrOfChannels = atoi(arg); argUsed = 1;
          if(Arguments->nrOfChannels > 2) {
            fprintf(stderr, "%s: -n %s: not %d..%d\n",
                    programName, arg, 0, 2);
            err = 1;
          }
          break;
        case 'C':  Arguments->Output24  = TRUE; break;
        case 'A':  Arguments->need_aiff = TRUE; break;
        case 'E':  Arguments->need_esps = TRUE; break;	/* MI */
        case 'R':  Arguments->need_raw  = TRUE; break;
#ifdef USELIB_AFsp
        case 'W':  Arguments->need_wav  = TRUE; break;
#endif
        case 'a':  Arguments->antialias_mixed_blocks = 0; break;
        default:   fprintf(stderr,"%s: unrecognized option %c\n",
                           programName, c);
        err = 1; break;
        }
        if(argUsed) {
          if(arg == token) token = ""; /* no more from token */
          else             ++i; /* skip arg we used */
          arg = ""; argUsed = 0;
        }
      }
    }
    else {
      if(Arguments->encoded_file_name[0] == '\0') {
        if ( strlen(argv[i]) >= MAX_NAME_SIZE ) {
          fprintf (stderr,"Input file path and name too long:\n%s\n\n", argv[i]);
          return(COMMAND_LINE_ERROR);
        }
        strncpy(Arguments->encoded_file_name, argv[i],MAX_NAME_SIZE);
      }
      else {
        if(Arguments->decoded_file_name[0] == '\0') {
          if ( strlen(argv[i]) >= MAX_NAME_SIZE ) {
            fprintf (stderr,"Output file path and name too long:\n%s\n\n", argv[i]);
            return(COMMAND_LINE_ERROR);
          }
          strncpy(Arguments->decoded_file_name, argv[i],MAX_NAME_SIZE);
        }
        else {
          if(Arguments->mps_file_name[0] == '\0') {
            if ( strlen(argv[i]) >= MAX_NAME_SIZE ) {
              fprintf (stderr,"MPEG Surround data file path and name too long:\n%s\n\n", argv[i]);
              return(COMMAND_LINE_ERROR);
            }
            strncpy(Arguments->mps_file_name, argv[i],MAX_NAME_SIZE);
          }
          else {
            fprintf(stderr,"%s: excess arg %s\n", programName, argv[i]);
            err = 1;
          }
        }
      }
    }
  }
  if(err || Arguments->encoded_file_name[0] == '\0') {
    usage(); 
    return (COMMAND_LINE_ERROR);
  }
      
  if(Arguments->decoded_file_name[0] == '\0') {
    strcpy(Arguments->decoded_file_name, Arguments->encoded_file_name);
    strcat(Arguments->decoded_file_name, DFLT_OPEXT);
  }
      
  if(Arguments->mps_file_name[0] == '\0') {
    strcpy(Arguments->mps_file_name, Arguments->encoded_file_name);
    strcat(Arguments->mps_file_name, DFLT_MPSEXT);
  }
      
    
#ifdef USELIB_AFsp
  if ( ( Arguments->need_aiff + Arguments->need_esps + Arguments->need_raw + Arguments->need_wav ) > 1 ) {
    printf("multiple output formats selected\n");
    return(COMMAND_LINE_ERROR);
  }
  if ( ( Arguments->need_aiff + Arguments->need_esps + Arguments->need_raw + Arguments->need_wav ) == 0 ) {
    Arguments->need_wav = 1;
  }
#else
  if ( ( Arguments->need_aiff + Arguments->need_esps + Arguments->need_raw ) > 1 ) {
    printf("multiple output formats selected\n");
    return(COMMAND_LINE_ERROR);
  }
  if ( ( Arguments->need_aiff + Arguments->need_esps + Arguments->need_raw ) == 0 ) {
    Arguments->need_raw = 1;
  }
#endif
  /* report results of dialog / command line */
  printf("Input file = '%s' output file = '%s' mps file = '%s'\n",
         Arguments->encoded_file_name, Arguments->decoded_file_name, Arguments->mps_file_name);
  if(Arguments->need_aiff) printf("Output file written in AIFF format\n");
  if(Arguments->need_esps) printf("Output file written in ESPS format\n"); /* MI */
  return (NO_ERROR);
}

unsigned int scalefac_IIP_buffer[54];	/* Intensity Stereo Bug 06.11.01 Vo */

/* Implementations */

int main( int argc, char** argv )
{
  Bit_stream_struc  bs = {NULL,NULL,0,0,0,0,0,0,0,0};
  frame_params      fr_ps;
  layer             info;
  unsigned short    init = 0;
  unsigned long     sample_frames;
  int               i, j, k, stereo=0, done=FALSE, sync; 
  unsigned short    clip;
  int               error_protection, crc_error_count, total_error_count;
  unsigned int      old_crc, new_crc;
  unsigned int      bit_alloc[2][SBLIMIT], scfsi[2][SBLIMIT];
  unsigned int      scale_index[2][3][SBLIMIT];
  unsigned long     bitsPerSlot=0, samplesPerFrame=0, frameNum = 0;
  unsigned long     frameBits, gotBits = 0;
  int               freeFrameSlots = 0;
  int               maxFrameSlots = 0;
  unsigned long	    longSync;
  Arguments_t       Arguments;
  int               Max_gr;
  ERROR_CODE        errorCode;
  unsigned int      implicitChannelReConfig = 0;
  int               anc_nbytes, anc_nbits;
  unsigned char     anc_bytes[2048];
  int               mps_pos, mps_len;
  static int        first = 1, skip_bytes = 0;
#ifdef PrintAncData
  long              logging_on = 1;
#else
  long              logging_on = 0;
#endif
    
  III_scalefac_t    III_scalefac;
  III_side_info_t   III_side_info;

  PCM FAR *pcm_sample = NULL;
  SAM FAR *sample     = NULL;
  FRA FAR *fraction   = NULL;
  VE  FAR *w          = NULL;

#ifdef USELIB_AFsp
  AFILE             *musicout = NULL;
  unsigned short    dFormat = FD_INT16;
  unsigned short    fType   = FTW_WAVE;
#else
  FILE              *musicout = NULL;
  IFF_AIFF          pcm_aiff_data;
#endif
  FILE              *mps_out = NULL;

#ifdef  MACINTOSH
  console_options.nrows = MAC_WINDOW_SIZE;
  argc = ccommand(&argv);
#endif
  
  /* Most large variables are declared dynamically to ensure
     compatibility with smaller machines */

  if ( ( pcm_sample = (PCM FAR *) mem_alloc((long) sizeof(PCM), "PCM Samp") ) == NULL ) {
    MyExit(bs,
           musicout,
           *pcm_sample,
           *sample,
           *fraction,
           *w,
           MEMORY_ALLOCATION_ERROR);
  }
  if ( ( sample     = (SAM FAR *) mem_alloc((long) sizeof(SAM), "Sample") )   == NULL ) {
    MyExit(bs,
           musicout,
           *pcm_sample,
           *sample,
           *fraction,
           *w,
           MEMORY_ALLOCATION_ERROR);
  }
  if ( ( fraction   = (FRA FAR *) mem_alloc((long) sizeof(FRA), "fraction") ) == NULL ) {
    MyExit(bs,
           musicout,
           *pcm_sample,
           *sample,
           *fraction,
           *w,
           MEMORY_ALLOCATION_ERROR);
  }
  if ( ( w          = (VE FAR  *) mem_alloc((long) sizeof(VE) , "w") )        == NULL ) {
    MyExit(bs,
           musicout,
           *pcm_sample,
           *sample,
           *fraction,
           *w,
           MEMORY_ALLOCATION_ERROR);
  }

  fr_ps.header = &info;
  fr_ps.tab_num = -1;                /* no table loaded */
  fr_ps.alloc = NULL;
  for (i=0;i<HAN_SIZE;i++) for (j=0;j<2;j++) (*w)[j][i] = 0.0;

  if (( errorCode = GetArguments(argc, argv, &Arguments) ) != NO_ERROR) {
    MyExit(bs,
           musicout,
           *pcm_sample,
           *sample,
           *fraction,
           *w,
           errorCode);
  }
;
#ifndef USELIB_AFsp
  if ((musicout = fopen(Arguments.decoded_file_name, "w+b")) == NULL) {
    printf ("Could not create \"%s\".\n", Arguments.decoded_file_name);
    MyExit(bs,
           musicout,
           *pcm_sample,
           *sample,
           *fraction,
           *w,
           FILE_OPEN_ERROR);
  }
#endif
  if ((mps_out = fopen(Arguments.mps_file_name, "w+b")) == NULL) {
    printf ("Could not create \"%s\".\n", Arguments.mps_file_name);
    MyExit(bs,
           musicout,
           *pcm_sample,
           *sample,
           *fraction,
           *w,
           FILE_OPEN_ERROR);
  }
  if ( ( errorCode = open_bit_stream_r(&bs, Arguments.encoded_file_name, BUFFER_SIZE) ) != NO_ERROR ) {
      MyExit(bs,
             musicout,
             *pcm_sample,
             *sample,
             *fraction,
             *w,
             errorCode);
  }
#ifndef USELIB_AFsp
  if (Arguments.need_aiff)
    if ( ( errorCode = aiff_seek_to_sound_data(musicout) ) != NO_ERROR) {
      printf("Could not seek to PCM sound data in \"%s\".\n",
             Arguments.decoded_file_name);
      MyExit(bs,
             musicout,
             *pcm_sample,
             *sample,
             *fraction,
             *w,
             errorCode);
    }
#endif

  sample_frames = 0;
  total_error_count = 0;  /* moved here from inside while() loop, efs 2003-11-11 */

/* from here several lines are added for FREE FORMAT by efs 2003-11-13 */
/* first examine the bit stream and look whether a syncword can be found */

  sync = seek_sync(&bs, SYNC_WORD, SYNC_WORD_LNGTH);
 
  if (!sync) {
      printf("Frame cannot be located\n");
      printf("Input stream may be empty\n");
      done = TRUE;
      MyExit(bs,
             musicout,
             *pcm_sample,
             *sample,
             *fraction,
             *w,
             errorCode);
  }
  else
  {
      freeFrameSlots = bs.buf_byte_idx;			/* save the index */
      decode_info(&bs, &fr_ps);
/* in case of free format, look for the second syncword */
/* and determine the corresponding frame length         */
	  if (fr_ps.header->bitrate_index == 0) {   
          printf("Free Format indicated");
/* it is not sufficient to simply search for the 12 bit "FFF" syncword like in:                */
/*          sync = seek_sync(&bs, SYNC_WORD, SYNC_WORD_LNGTH);                                 */
/* This pattern is often present in the actual bit stream. Therefore the first 20 bits         */
/* of the header are reconstructed and then we search for that longer pattern.                 */
/* This is not without problems, as                                                            */
/*    1.  this pattern may also be emulated by the bit stream                                  */
/*    2.  it is not excluded by the standard to change the crc protection from frame to frame. */
/* Nevertheless, this code is a useful example and allows to decode FREE FORMAT.               */
	
#ifdef MPEG_2_5
          longSync = (((SYNC_WORD << 2) | fr_ps.header->version) << 2) | (fr_ps.header->lay);
#else
	  longSync = (((SYNC_WORD << 1) | fr_ps.header->version) << 2) | (4-fr_ps.header->lay);
#endif
	  longSync = (((longSync << 1) | !fr_ps.header->error_protection) << 4) | fr_ps.header->bitrate_index;
          sync = seek_sync(&bs, longSync, 20);
          if (!sync) {
              printf(", but no 2nd frame can be located\n");
              done = TRUE;
              MyExit(bs,
                     musicout,
                     *pcm_sample,
                     *sample,
                     *fraction,
                     *w,
                     errorCode);
		  }
          else
		  {
			freeFrameSlots = freeFrameSlots - (bs.buf_byte_idx + 1);     /* the 2nd sync word is 1 slot longer */
			if (freeFrameSlots <= 0) freeFrameSlots += (bs.buf_size - MINIMUM);  /* Buffer is actually shorter than buf_size */
			printf(" with %d slots per frame\n",freeFrameSlots);

			/* get number of slots for maximum supported bit rate, depending on version, layer, fs */
            maxFrameSlots = (int)((144 * bitrate[fr_ps.header->version][fr_ps.header->lay - 1][14])
              / s_freq[fr_ps.header->version][fr_ps.header->sampling_frequency]);

			if (freeFrameSlots > maxFrameSlots)
				printf("WARNING: decoder does not have to support more than %d slots per frame.\n",maxFrameSlots);
		  }
	  }
  }
  rewind_bit_stream_r(&bs);

  while (!end_bs(&bs)) {
    sync = seek_sync(&bs, SYNC_WORD, SYNC_WORD_LNGTH);
    frameBits = sstell(&bs) - gotBits;
    if(frameNum > 0)        /* don't want to print on 1st loop; no lay */
      if(frameBits%bitsPerSlot)
        fprintf(stderr,"Got %ld bits = %ld slots plus %ld\n",
                frameBits, frameBits/bitsPerSlot, frameBits%bitsPerSlot);
    gotBits += frameBits;
    anc_nbytes = 0;    

    if (!sync) {
      printf("Frame cannot be located\n");
      printf("Input stream may be empty\n");
      done = TRUE;
      break;
    }

    decode_info(&bs, &fr_ps);
    hdr_to_frps(&fr_ps);
    if ( ! init ) {
      init = 1;
      if (Arguments.nrOfChannels) {
        fr_ps.stereoInit = Arguments.nrOfChannels;
      }
      else {
        fr_ps.stereoInit = fr_ps.stereo;
      }
#ifdef USELIB_AFsp
      if (Arguments.Output24) {
        dFormat = FD_INT24;
      }
      if (Arguments.need_aiff) {
        fType = FTW_AIFF;
      }
      else if ( Arguments.need_raw ) {
        fType = FTW_NH_EB;
      }
       (AFoptions ())->ErrorHalt = 0;
      if ((musicout = AFopenWrite ( Arguments.decoded_file_name, 
                                    dFormat+256*fType,
                                    fr_ps.stereoInit,
                                    s_freq[info.version][info.sampling_frequency]*1000,
                                    NULL )) == NULL ) {
        printf ("Could not create \"%s\".\n", Arguments.decoded_file_name);
        MyExit(bs,
               musicout,
               *pcm_sample,
               *sample,
               *fraction,
               *w,
               FILE_OPEN_ERROR);
      }
#endif
    }
#ifdef CHANNEL_INVESTIGATION
    printf("channels:%d\n",fr_ps.stereo );
#endif /* CHANNEL_INVESTIGATION */
    if ( ! implicitChannelReConfig && fr_ps.stereoInit < fr_ps.stereo ) {
      implicitChannelReConfig = 1;
      fprintf (stderr, "\n\
WARNING: implicit channel re-configuration\n\
             + initial number of channels: %d\n\
             + current number of channels: %d\n",
               fr_ps.stereoInit,fr_ps.stereo);
      if ( ! Arguments.need_raw ) {
      fprintf (stderr, "\
             --> this is not supported by the selected output file format\n\
             --> decoding stopped\n\n");
        MyExit(bs,
               musicout,
               *pcm_sample,
               *sample,
               *fraction,
               *w,
               CHANNEL_CONFIG_ERROR);
      }
    }
    stereo = fr_ps.stereo;
#ifdef MPEG_2_5
    if(fr_ps.header->version != MPEG_AUDIO_ID)
#else
    if(fr_ps.header->version == MPEG_PHASE2_LSF) 
#endif
    {
      Max_gr = 1;
    }
    else { 
      Max_gr = 2;
    }

    error_protection = info.error_protection;
    crc_error_count = 0;
    /* total_error_count = 0;  removed to outside while() loop, efs 2003-11-11 */
    if(frameNum == 0) WriteHdr(&fr_ps, stdout);  /* printout layer/mode */ 

#ifdef ESPS
    if (frameNum == 0 && Arguments.need_esps) {
      esps_write_header(musicout,(long) sample_frames, (double)
                        s_freq[info.version][info.sampling_frequency] * 1000,
                        (int) stereo, Arguments.decoded_file_name );
    } /* MI */
#endif

    fprintf(stderr, "\r[%4lu]", frameNum++); fflush(stderr); 
    if (error_protection) buffer_CRC(&bs, &old_crc);

    switch (info.lay) {
#ifdef MPEG_2_5
	    case 3:
#else
	    case 1:
#endif
      bitsPerSlot = 32;        samplesPerFrame = 384;
      I_decode_bitalloc(&bs,bit_alloc,&fr_ps);
      I_decode_scale(&bs, bit_alloc, scale_index, &fr_ps);
      
      if (error_protection) {
        I_CRC_calc(&fr_ps, bit_alloc, &new_crc);
        if (new_crc != old_crc) {
          crc_error_count++;
          total_error_count++;
#ifndef USELIB_AFsp 
          recover_CRC_error(*pcm_sample, crc_error_count,
                            &fr_ps, musicout, &sample_frames, Arguments.Output24);
#endif
          break;
        }
        else crc_error_count = 0;
      }
      
      clip = 0;
      for (i=0;i<SCALE_BLOCK;i++) {
        I_buffer_sample(&bs,(*sample),bit_alloc,&fr_ps);
        I_dequantize_sample(*sample,*fraction,bit_alloc,&fr_ps);
        I_denormalize_sample((*fraction),scale_index,&fr_ps);
        
        if(Arguments.topSb>0)        /* clear channels to 0 */
          for(j=Arguments.topSb; j<fr_ps.sblimit; ++j)
            for(k=0; k<stereo; ++k)
              (*fraction)[k][0][j] = 0;
        
        for (j=0;j<stereo;j++) {
          if ( ( errorCode = SubBandSynthesis (&((*fraction)[j][0][0]), j,
                                    &((*pcm_sample)[j][0][0])) ) != NO_ERROR ) {
          MyExit(bs,
                 musicout,
                 *pcm_sample,
                 *sample,
                 *fraction,
                 *w,
                 errorCode);
           }
        }
        if ( ( errorCode = out_fifo(*pcm_sample, 1, &fr_ps, done,
                      musicout, &sample_frames, Arguments.Output24, &clip) ) != NO_ERROR ) {
          MyExit(bs,
                 musicout,
                 *pcm_sample,
                 *sample,
                 *fraction,
                 *w,
                 errorCode);
        }
      }
      if(clip > 0) fprintf(stderr," %3d output samples clipped\n", clip);
      break;
      
    case 2 : 
      bitsPerSlot = 8;        
      samplesPerFrame = 1152;
      II_decode_bitalloc(&bs, bit_alloc, &fr_ps);
      II_decode_scale(&bs, scfsi, bit_alloc, scale_index, &fr_ps);

      if (error_protection) { 
        II_CRC_calc(&fr_ps, bit_alloc, scfsi, &new_crc);
        if (new_crc != old_crc) {
          crc_error_count++;
          total_error_count++;
#ifndef USELIB_AFsp
          recover_CRC_error(*pcm_sample, crc_error_count,
                            &fr_ps, musicout, &sample_frames, Arguments.Output24);
#endif
          break;
        }
        else crc_error_count = 0;
      }

      clip = 0;
      for (i=0;i<SCALE_BLOCK;i++) {
        II_buffer_sample(&bs,(*sample),bit_alloc,&fr_ps);
        II_dequantize_sample((*sample),bit_alloc,(*fraction),&fr_ps);
        II_denormalize_sample((*fraction),scale_index,&fr_ps,i>>2);

        if(Arguments.topSb>0)        /* debug : clear channels to 0 */
          for(j=Arguments.topSb; j<fr_ps.sblimit; ++j)
            for(k=0; k<stereo; ++k)
              (*fraction)[k][0][j] =
                (*fraction)[k][1][j] =
                (*fraction)[k][2][j] = 0;

        for (j=0;j<3;j++) for (k=0;k<stereo;k++) {
          if ( ( errorCode = SubBandSynthesis (&((*fraction)[k][j][0]), k,
                                               &((*pcm_sample)[k][j][0])) ) != NO_ERROR ) {
            MyExit(bs,
                   musicout,
                   *pcm_sample,
                   *sample,
                   *fraction,
                   *w,
                   errorCode);
          }
        }
        if ( (errorCode = out_fifo(*pcm_sample, 3, &fr_ps, done, musicout,
                                   &sample_frames, Arguments.Output24, &clip) ) != NO_ERROR ) {
          MyExit(bs,
                 musicout,
                 *pcm_sample,
                 *sample,
                 *fraction,
                 *w,
                 errorCode);
        }
      }
      if(clip > 0) fprintf(stderr," %3d samples clipped\n", clip);
      break;

#ifdef MPEG_2_5
    case 1:
#else
    case 3:
#endif
      {
        int nSlots;
        int gr, ch, ss, sb, main_data_end, flush_main ;
        int  bytes_to_discard ;
        static int frame_start = 0;
        
        bitsPerSlot = 8;
#ifdef MPEG_2_5
	if (fr_ps.header->version != MPEG_AUDIO_ID)
#else
        if(fr_ps.header->version == MPEG_PHASE2_LSF)
#endif
	{
          samplesPerFrame = 576;
	}
        else
	{
          samplesPerFrame = 1152;
	  }
        
/* CRC calculation added to III_get_side_info() by efs 2003-11-11 */
/* done in any case, actually not needed if error_protection == false */
        errorCode = III_get_side_info_crc(&bs, &III_side_info, &fr_ps, &new_crc);
        if (errorCode != NO_ERROR) {
          MyExit(bs,
                 musicout,
                 *pcm_sample,
                 *sample,
                 *fraction,
                 *w,
                 errorCode);
        }
/* CRC evaluation and error printout, added by efs 2003-11-11 */
		if (error_protection) { 
			if (new_crc != old_crc) {
			    crc_error_count++;
			    total_error_count++;
                fprintf(stderr," CRC error\n");
			}
			else crc_error_count = 0;
		}

        nSlots = main_data_slots(fr_ps);
	if (fr_ps.header->bitrate_index == 0) nSlots += freeFrameSlots;  /* efs 2003-11-13 */
        
	for (; nSlots > 0; nSlots--)  /* read main data. */
          hputbuf((unsigned int) getbits(&bs,8), 8);
        main_data_end = hsstell() / 8; /*of privious frame*/
        if ( (flush_main=(hsstell() % bitsPerSlot))!=0 ) { 
          hgetbits((int)(bitsPerSlot - flush_main));
          main_data_end ++;
        }
        bytes_to_discard = frame_start - main_data_end
          - III_side_info.main_data_begin ;
        if( main_data_end > 4096 )
          {   frame_start -= 4096;
          rewindNbytes( 4096 );
          }
        
        frame_start += main_data_slots(fr_ps);
		if (fr_ps.header->bitrate_index == 0) frame_start += freeFrameSlots;  /* efs 2003-11-13 */
        if (bytes_to_discard < 0) {
          printf("Not enough main data to decode frame %ld.  Frame discarded.\n", 
                 frameNum - 1); break;
        }
#ifdef PrintAncDataOld
        printf ("bytes_to_discard=%04d ", bytes_to_discard);
#endif
        for (; bytes_to_discard > 0; bytes_to_discard--)
        {
          int byte = hgetbits(8);
          anc_bytes[anc_nbytes++] = byte;
#ifdef PrintAncDataOld
          printf (" %02X", byte);
#endif
        }
#ifdef PrintAncDataOld
        printf ("\n");
        fflush (stdout);
#endif

        
        clip = 0;
        for (gr=0;gr<Max_gr;gr++) {
          double lr[2][SBLIMIT][SSLIMIT],ro[2][SBLIMIT][SSLIMIT];
          
          for (ch=0; ch<stereo; ch++) {
            long int is[SBLIMIT][SSLIMIT];   /* Quantized samples. */ 
            int part2_start;
            part2_start = hsstell();
#ifdef MPEG_2_5
	    if(fr_ps.header->version == MPEG_AUDIO_ID)
#else
            if(fr_ps.header->version != MPEG_PHASE2_LSF)
#endif
	    {
                III_get_scale_factors(&III_scalefac,&III_side_info,gr,ch);
              }
            else {
                III_get_LSF_scale_factors(&III_scalefac,
                                          &III_side_info,
                                          gr,ch,&fr_ps);
              }
            if ( ( errorCode = III_huffman_decode(is, &III_side_info, ch, gr, part2_start,
                                                 &fr_ps) ) != NO_ERROR ) {
              MyExit(bs,
                     musicout,
                     *pcm_sample,
                     *sample,
                     *fraction,
                     *w,
                     errorCode);
             }
            III_dequantize_sample(is, ro[ch], &III_scalefac,
                                  &(III_side_info.ch[ch].gr[gr]), ch, &fr_ps);
          }
#ifdef MPEG_2_5
	    if(fr_ps.header->version == MPEG_AUDIO_ID)
#else
            if(fr_ps.header->version != MPEG_PHASE2_LSF)
#endif
	   {
            MPEG_1_III_stereo(ro, lr, &III_scalefac, &(III_side_info.ch[0].gr[gr]), &fr_ps);
	   }
          else 
	  {
	    MPEG_2_III_stereo(ro, lr, &III_scalefac, &(III_side_info.ch[0].gr[gr]), &(III_side_info.ch[1].gr[gr]), &fr_ps);
	  }
          
          for (ch=0; ch<stereo; ch++) { 
            double re[SBLIMIT][SSLIMIT];
            double hybridIn[SBLIMIT][SSLIMIT];/* Hybrid filter input */
            double hybridOut[SBLIMIT][SSLIMIT];/* Hybrid filter out */
            double polyPhaseIn[SBLIMIT];     /* PolyPhase Input. */
            
            III_reorder (lr[ch],re,&(III_side_info.ch[ch].gr[gr]),
                         &fr_ps);
            
            III_antialias(re, hybridIn, 
                          &(III_side_info.ch[ch].gr[gr]),
                          Arguments.antialias_mixed_blocks);
            
	    for (sb=0; sb<SBLIMIT; sb++) { /* Hybrid synthesis. */
              III_hybrid(hybridIn[sb], hybridOut[sb], sb, ch,&(III_side_info.ch[ch].gr[gr]));
            }
            
            for (ss=0;ss<18;ss++) /*Frequency inversion for polyphase.*/
              for (sb=0; sb<SBLIMIT; sb++)
                if ((ss%2) && (sb%2))
                  hybridOut[sb][ss] = -hybridOut[sb][ss];
            
            for (ss=0;ss<18;ss++) { /* Polyphase synthesis */
              for (sb=0; sb<SBLIMIT; sb++)
                polyPhaseIn[sb] = hybridOut[sb][ss];
              if ( ( errorCode = SubBandSynthesis (polyPhaseIn, ch, &((*pcm_sample)[ch][ss][0])) ) != NO_ERROR ) {
                MyExit(bs,
                       musicout,
                       *pcm_sample,
                       *sample,
                       *fraction,
                       *w,
                       errorCode);
              }
            }
          }
          /* Output PCM sample points for one granule. */
          if ( ( errorCode = out_fifo(*pcm_sample, 18, &fr_ps, done, musicout,
                                       &sample_frames, Arguments.Output24, &clip) ) != NO_ERROR ) {
            MyExit(bs,
                   musicout,
                   *pcm_sample,
                   *sample,
                   *fraction,
                   *w,
                   errorCode);
          }
        }
        if(clip > 0) fprintf(stderr," %3d samples clipped.\n", clip);
        break;
      }
    }

    /* skip ancillary data   HP 22-nov-95 */
    if (info.bitrate_index > 0) { /* if not free-format */
      long anc_len;
      anc_len = (int)((double)samplesPerFrame /
                      s_freq[info.version][info.sampling_frequency] *
#ifdef MPEG_2_5
                      (double)bitrate[info.version][info.lay][info.bitrate_index] /
#else
		      (double)bitrate[info.version][info.lay-1][info.bitrate_index] /
#endif
                      (double)bitsPerSlot);
      if (info.padding)
        anc_len++;
      anc_len *= bitsPerSlot;
         
#ifdef MPEG_2_5
      anc_len -= sstell(&bs)-gotBits+SYNC_WORD_LNGTH;
#else
      anc_len -= sstell(&bs)-gotBits+SYNC_WORD_LNGTH;
#endif

      anc_nbits = anc_len % 8;
      for (j=0; j<anc_nbits; j++) get1bit(&bs);
      anc_len -= anc_nbits;

      for (; anc_len > 0; anc_len -= 8)
      {
        int byte = getbits(&bs, 8);
        anc_bytes[anc_nbytes++] = byte;
      }
#ifdef PrintAncData
      printf ("anc_data: ");
#endif
      for (j=0; j<anc_nbytes; j++)
      {
#ifdef PrintAncData
        printf (" %02X", anc_bytes[j]);
#endif
      }
#ifdef PrintAncData
      printf ("\n");
      fflush (stdout);
#endif

      mps_len = find_mps_data (anc_bytes, anc_nbytes, 0, &mps_pos, &first, &skip_bytes, logging_on);
      if (mps_len > 0)
        fwrite (anc_bytes+mps_pos, 1, mps_len, mps_out);
    }
  }
#ifdef MPEG_2_5
  if ((info.lay == 1) && (mps_out != NULL))
#else
  if ((info.lay == 3) && (mps_out != NULL))
#endif
  {
    anc_nbytes = hsstell() / 8;
    if (anc_nbytes > 2048) {
      anc_nbytes = 2048;
    }
    for (j=0; j<anc_nbytes; j++)
    {
      anc_bytes[j] = (unsigned char) hgetbits(8);
      mps_len = find_mps_data (anc_bytes, j+1, 0, &mps_pos, &first, &skip_bytes, logging_on);
      if (mps_len > 0)
      {
        fwrite (anc_bytes+mps_pos, 1, mps_len, mps_out);
        break;
      }
    }
  }

  /* finally write out the (last incomplete) buffer */
  done=TRUE;
  if ( ( errorCode =  out_fifo(*pcm_sample, 0, &fr_ps, done,
                               musicout, &sample_frames, Arguments.Output24, &clip ) ) != NO_ERROR ) {
    MyExit(bs,
           musicout,
           *pcm_sample,
           *sample,
           *fraction,
           *w,
           errorCode);
  }
  if (clip > 0) fprintf(stderr," %3d samples clipped.\n", clip);
  
#ifndef USELIB_AFsp
  if (Arguments.need_aiff) {
    pcm_aiff_data.numChannels       = fr_ps.stereoInit;
    pcm_aiff_data.numSampleFrames   = sample_frames;
    if (Arguments.Output24) { 
      pcm_aiff_data.sampleSize        = 24; 
    } 
    else { 
      pcm_aiff_data.sampleSize        = 16;
    } 
    pcm_aiff_data.sampleRate        = s_freq[info.version][info.sampling_frequency]*1000;
    pcm_aiff_data.sampleType        = IFF_ID_SSND;
    pcm_aiff_data.blkAlgn.offset    = 0;
    pcm_aiff_data.blkAlgn.blockSize = 0;

    if ( ( errorCode = aiff_write_headers(musicout, &pcm_aiff_data) ) != NO_ERROR) {
      printf("Could not write AIFF headers to \"%s\"\n",
             Arguments.decoded_file_name);
      MyExit(bs,
             musicout,
             *pcm_sample,
             *sample,
             *fraction,
             *w,
             errorCode);
    }
  }
#endif

  printf("Avg slots/frame = %.3f; b/smp = %.2f; br = %.3f kbps\n",
         (FLOAT) gotBits / (frameNum * bitsPerSlot),
         (FLOAT) gotBits / (frameNum * samplesPerFrame),
         (FLOAT) gotBits / (frameNum * samplesPerFrame) *
         s_freq[info.version][info.sampling_frequency]);

  /* tell how many CRC errors, if any.  Added by efs 2003-11-11 */
  if (total_error_count > 0) printf("Total %d CRC error(s).\n", total_error_count);
  
  MyClose(bs,
          musicout,
          *pcm_sample,
          *sample,
          *fraction,
          *w);
  if (mps_out != NULL) fclose (mps_out);
  /* for the correct AIFF header information */
  /*             on the Macintosh            */
  /* the file type and the file creator for  */
  /* Macintosh compatible Digidesign is set  */

#ifdef  MACINTOSH
  if (Arguments.need_aiff)
    set_mac_file_attr(Arguments.decoded_file_name, VOL_REF_NUM,
                      CREATR_DEC_AIFF, FILTYP_DEC_AIFF);
  else	set_mac_file_attr(Arguments.decoded_file_name, VOL_REF_NUM,
                          CREATR_DEC_BNRY, FILTYP_DEC_BNRY);
#endif

  printf("Decoding of \"%s\" is finished\n", Arguments.encoded_file_name);
  printf("The decoded PCM output file name is \"%s\"\n", Arguments.decoded_file_name);
  if (Arguments.need_aiff) {
    printf("\"%s\" has been written with AIFF header information\n",
           Arguments.decoded_file_name);
  }
  if (Arguments.need_wav) {
    printf("\"%s\" has been written with WAVE header information\n",
           Arguments.decoded_file_name);
  }
  return ( 0 );
}

